from flask import Flask, request
from flask_cache import Cache
from flask_restful import Resource, Api, reqparse
from speculator import market
from webargs import fields
from webargs.flaskparser import use_kwargs

app = Flask(__name__)
api = Api(app)

# TODO: Add private API with Redis
cache = Cache(app, config={'CACHE_TYPE': 'simple'})
cache.init_app(app)

@api.resource('/api/public/predict')
class predict(Resource):
    """ Predict the next price of a symbol like USDT_BTC """
    # TODO: Add private POST/PUT/DELETE methods

    @use_kwargs({
        'model': fields.Str(missing='rf'),
        'symbol': fields.Str(missing='USDT_BTC'),
        'unit': fields.Str(missing='month'),
        'count': fields.Int(missing=6),
        'period': fields.Integer(missing=86400),
        'partition': fields.Integer(missing=14),
        'delta': fields.Integer(missing=25),
        'seed': fields.Integer(missing=None),
        'trees': fields.Integer(missing=10),
        'jobs': fields.Integer(missing=1),
        'longs': fields.DelimitedList(fields.Str(), missing=[])
    })
    @cache.memoize(3600)
    def get(self, model, symbol, unit, count, period,
            partition, delta, seed, trees, jobs, longs):
        m = market.Market(symbol=symbol, unit=unit,
                          count=count, period=period)

        features = m.set_features(partition=partition)
        features = m.set_long_features(features,
                                       columns_to_set=longs,
                                       partition=partition)

        targets = market.set_targets(features, delta=delta)
        features = features.drop(['close'], axis=1)

        model = market.setup_model(features[:-1], targets,
                                   model_type=model.lower(),
                                   seed=seed,
                                   n_estimators=trees,
                                   n_jobs=jobs)

        next_date = features.tail(1) # Remember the entry we didn't train?  Predict it.

        trend = market.target_code_to_name(model._predict_trends(next_date)[0])
        accuracy = model.accuracy(model.features.test, model.targets.test)
        proba = model._predict_probas(next_date)
        proba_log = model._predict_logs(next_date) # Logarithmic scale

        return {
            "trend": trend,
            "test_set_accuracy": "{0:.3f}%".format(100 * accuracy),
            "probabilities": {
                market.target_code_to_name(code): p for code, p in enumerate(proba[0])
            }
        }

if __name__ == '__main__':
    app.run(debug=True)



from collections import namedtuple
from datetime import datetime as dt
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split
from speculator.features.OBV import OBV
from speculator.features.RSI import RSI
from speculator.features.SMA import SMA
from speculator.features.SO import SO
import speculator.models.random_forest as rf
import speculator.models.deep_neural_network as dnn
from speculator.utils import date
from speculator.utils import poloniex

# Enum-like object with 1:1 mapping.  Converts a readable
# market trend like 'bearish' to an int that is easier to parse.
TARGET_CODES = {'bearish': 0, 'neutral': 1, 'bullish': 2}

class Market:
    """ Evaluates TA indicators of a market

    Gets data from a market and then calculates technical analysis features.
    It also prepares this data to be fed into machine learning interfaces
    by creating a pandas DataFrame, and generating training/test sets
    for features (x-axis) and the target market trend (y-axis).

    Attributes:
        symbol: String of currency pair, like a ticker symbol.
        unit: String of time period unit for count argument.
            How far back to check historical market data.
            valid values: 'hour', 'day', 'week', 'month', 'year'
        count: Int of units.
            How far back to check historical market data.
        period: Int defining width of each chart candlestick in seconds.
            Valid values: 300, 900, 1800, 7200, 14400, 86400.
        json: JSON data as a list of dict dates, where the keys are
            the raw market statistics.
    """

    def __init__(self, json=None, symbol='USDT_BTC', unit='month', count=6, period=86400):
        """ Inits market class of symbol with data going back count units """
        self.symbol = symbol
        self.unit = unit
        self.count = count
        self.period = period

        if json is None:
            self.json = self.get_json()
        else:
            self.json = json

    def get_json(self):
        """ Gets market chart data from today to a previous date """
        today = dt.now()
        DIRECTION = 'last'
        epochs = date.get_end_start_epochs(today.year, today.month, today.day,
                                           DIRECTION, self.unit, self.count)
        return poloniex.chart_json(epochs['shifted'], epochs['initial'],
                                   self.period, self.symbol)[0]

    def set_features(self, partition=1):
        """ Parses market data JSON for technical analysis indicators

        Args:
            partition: Int of how many dates to take into consideration
                when evaluating technical analysis indicators.

        Returns:
            Pandas DataFrame instance with columns as numpy.float32 features.
        """
        if len(self.json) < partition + 1:
            raise ValueError('Not enough dates for the specified partition size: {0}.  Try a smaller partition.'.format(partition))

        data = []
        for offset in range(len(self.json) - partition):
            json = self.json[offset : offset + partition]
            data.append(eval_features(json))
        return pd.DataFrame(data=data, dtype=np.float32)

    def set_long_features(self, features, columns_to_set=[], partition=2):
        """ Sets features of double the duration

        Example: Setting 14 day RSIs to longer will create add a
            feature column of a 28 day RSIs.

        Args:
            features: Pandas DataFrame instance with columns as numpy.float32 features.
            columns_to_set: List of strings of feature names to make longer
            partition: Int of how many dates to take into consideration
                when evaluating technical analysis indicators.

        Returns:
            Pandas DataFrame instance with columns as numpy.float32 features.
        """
        # Create long features DataFrame
        features_long = self.set_features(partition=2 * partition)

        # Remove features not specified by args.long
        unwanted_features = [f for f in features.columns if f not in columns_to_set]
        features_long = features_long.drop(unwanted_features, axis=1)

        # Prefix long columns with 'long_' to fix naming conflicts
        features_long.columns = ['long_{0}'.format(f) for f in features_long.columns]

        # Merge the two DataFrames
        skip = partition
        return pd.concat([features[skip:].reset_index(drop=True),
                         features_long],
                         axis=1)

def set_targets(x, delta=10):
    """ Sets target market trend for a date

    Args:
        x: Pandas DataFrame of market features
        delta: Positive number defining a price buffer between what is
            classified as a bullish/bearish market for the training set.
            delta is equivalent to the total size of the neutral price zone.
            delta / 2 is equivalent to either the positive or negative
            threshold of the neutral price zone.

    Returns:
        Pandas Series of numpy int8 market trend targets
    """
    data = [] # Keep track of targets
    for row, _ in x.iterrows():
        if row == x.shape[0] - 1: # Can't predict yet, done.
            break

        # Get closing prices
        curr_close = x.close[row]
        next_close = x.close[row + 1]
        high_close = next_close + (delta / 2) # Pos. neutral zone threshold
        low_close = next_close - (delta / 2)  # Neg. neutral zone threshold

        # Get target
        if curr_close < low_close:
            target = TARGET_CODES['bearish']
        elif curr_close > high_close:
            target = TARGET_CODES['bullish']
        else:
            target = TARGET_CODES['neutral']
        data.append(target)

    return pd.Series(data=data, dtype=np.int32, name='target')

def eval_features(json):
    """ Gets technical analysis features from market data JSONs

    Args:
        json: JSON data as a list of dict dates, where the keys are
            the raw market statistics.

    Returns:
        Dict of market features and their values
    """
    return {'close'    : json[-1]['close'],
            'sma'      : SMA.eval_from_json(json),
            'rsi'      : RSI.eval_from_json(json),
            'so'       : SO.eval_from_json(json),
            'obv'      : OBV.eval_from_json(json)}

def target_code_to_name(code):
    """ Converts an int target code to a target name

    Since self.TARGET_CODES is a 1:1 mapping, perform a reverse lookup
    to get the more readable name.

    Args:
        code: Value from self.TARGET_CODES

    Returns:
        String target name corresponding to the given code.
    """
    TARGET_NAMES = {v: k for k, v in TARGET_CODES.items()}
    return TARGET_NAMES[code]

def setup_model(x, y, model_type='random_forest', seed=None, **kwargs):
    """ Initializes a machine learning model

    Args:
        x: Pandas DataFrame, X axis of features
        y: Pandas Series, Y axis of targets
        model_type: Machine Learning model to use
            Valid values: 'random_forest'
        seed: Random state to use when splitting sets and creating the model
        **kwargs: Scikit Learn's RandomForestClassifier kwargs

    Returns:
        Trained model instance of model_type
    """
    assert len(x) > 1 and len(y) > 1, 'Not enough data objects to train on (minimum is at least two, you have (x: {0}) and (y: {1}))'.format(len(x), len(y))

    sets = namedtuple('Datasets', ['train', 'test'])
    x_train, x_test, y_train, y_test = train_test_split(x,
                                                        y,
                                                        random_state=seed,
                                                        shuffle=False)
    x = sets(x_train, x_test)
    y = sets(y_train, y_test)

    if model_type == 'random_forest' or model_type == 'rf':
        model = rf.RandomForest(x, y, random_state=seed, **kwargs)
    elif model_type == 'deep_neural_network' or model_type == 'dnn':
        model = dnn.DeepNeuralNetwork(x, y, **kwargs)
    else:
        raise ValueError('Invalid model type kwarg')
    return model



from speculator.features.TechnicalAnalysis import TechnicalAnalysis
from speculator.utils import poloniex

class OBV(TechnicalAnalysis):
    """
    On Balance Volume:
    OBV = OBV_prev + v
        such that,
            v = volume  if close >  close_prev
            v = 0       if close == close_prev
            v = -volume if close <  close_prev
    """

    def eval_algorithm(curr, prev):
        """ Evaluates OBV

        Args:
            curr: Dict of current volume and close
            prev: Dict of previous OBV and close

        Returns:
            Float of OBV
        """
        if curr['close'] > prev['close']:
            v = curr['volume']
        elif curr['close'] < prev['close']:
            v = curr['volume'] * -1
        else:
            v = 0
        return prev['obv'] + v

    def eval_from_json(json):
        """ Evaluates OBV from JSON (typically Poloniex API response)

        Args:
            json: List of dates where each entry is a dict of raw market data.

        Returns:
            Float of OBV
        """
        closes = poloniex.get_attribute(json, 'close')
        volumes = poloniex.get_attribute(json, 'volume')
        obv = 0
        for date in range(1, len(json)):
            curr = {'close': closes[date], 'volume': volumes[date]}
            prev = {'close': closes[date - 1], 'obv': obv}
            obv = OBV.eval_algorithm(curr, prev)
        return obv



from speculator.features.TechnicalAnalysis import TechnicalAnalysis
from speculator.utils import poloniex
from speculator.utils import stats

class SMA(TechnicalAnalysis):
    """
    Simple Moving Average:
    Average closing price over a period
    SMA = avg(closes) = sum(closes) / len(closes)
    """

    def eval_algorithm(closes):
        """ Evaluates the SMA algorithm

        Args:
            closes: List of price closes.

        Returns:
            Float average of closes.
        """
        return stats.avg(closes)

    def eval_from_json(json):
        """ Evaluates SMA from JSON (typically Poloniex API response)

        Args:
            json: List of dates where each entry is a dict of raw market data.

        Returns:
            Float average of closes.
        """
        closes = [date['close'] for date in json]
        return SMA.eval_algorithm(closes)



from speculator.features.TechnicalAnalysis import TechnicalAnalysis
from speculator.utils import poloniex
from speculator.utils import stats

class RSI(TechnicalAnalysis):
    """
    Relative Strength Index:
    RSI = 100 - (100 / (1 + RS))
        such that,
            RS = avg(t-period gain) / avg(t-period loss)
    """

    def eval_algorithm(gains, losses):
        """ Evaluates the RSI algorithm

        Args:
            gains: List of price gains.
            losses: List of prices losses.

        Returns:
            Float between 0 and 100, momentum indicator
            of a market measuring the speed and change of price movements.
        """
        return 100 - (100 / (1 + RSI.eval_rs(gains, losses)))

    def eval_rs(gains, losses):
        """ Evaluates the RS variable in RSI algorithm

        Args:
            gains: List of price gains.
            losses: List of prices losses.

        Returns:
            Float of average gains over average losses.
        """
        # Number of days that the data was collected through
        count = len(gains) + len(losses)

        avg_gains = stats.avg(gains, count=count) if gains else 1
        avg_losses = stats.avg(losses,count=count) if losses else 1
        if avg_losses == 0:
            return avg_gains
        else:
            return avg_gains / avg_losses

    def eval_from_json(json):
        """ Evaluates RSI from JSON (typically Poloniex API response)

        Args:
            json: List of dates where each entry is a dict of raw market data.

        Returns:
            Float between 0 and 100, momentum indicator
            of a market measuring the speed and change of price movements.
        """
        changes = poloniex.get_gains_losses(poloniex.parse_changes(json))
        return RSI.eval_algorithm(changes['gains'], changes['losses'])



import abc

class TechnicalAnalysis(abc.ABC):
    """ Abstract class for calculating technical analysis indicators """
    
    @staticmethod
    @abc.abstractmethod
    def eval_algorithm(*args, **kwargs):
        """ Evaluates TA algorithm """

    @staticmethod
    @abc.abstractmethod
    def eval_from_json(json):
        """ Evaluates TA algorithm from JSON

        Args:
            json: List of dates where each entry is a dict of raw market data.
        """



from speculator.features.SO import SO
import unittest

class SOTest(unittest.TestCase):
    def test_eval_algorithm(self):
        closing = 127.29
        low     = 124.56
        high    = 128.43
        self.assertAlmostEqual(SO.eval_algorithm(closing, low, high),
                               70.54, places=2)



from speculator.features.TechnicalAnalysis import TechnicalAnalysis
from speculator.utils import poloniex

class SO(TechnicalAnalysis):
    """
    Stochastic Oscillator:
    %K = 100 * (C - L(t)) / (H14 - L(t))
        such that,
            C = Current closing Price
            L(t) = Lowest Low over some duration t
            H(t) = Highest High over some duration t
              * t is usually 14 days

    %K follows the speed/momentum of a price in a market
    """

    def eval_algorithm(closing, low, high):
        """ Evaluates the SO algorithm

        Args:
            closing: Float of current closing price.
            low: Float of lowest low closing price throughout some duration.
            high: Float of highest high closing price throughout some duration.

        Returns:
            Float SO between 0 and 100.
        """
        if high - low == 0: # High and low are the same, zero division error
            return 100 * (closing - low)
        else:
            return 100 * (closing - low) / (high - low)

    def eval_from_json(json):
        """ Evaluates SO from JSON (typically Poloniex API response)

        Args:
            json: List of dates where each entry is a dict of raw market data.

        Returns:
            Float SO between 0 and 100.
        """
        close  = json[-1]['close'] # Latest closing price
        low    = min(poloniex.get_attribute(json, 'low')) # Lowest low
        high   = max(poloniex.get_attribute(json, 'high')) # Highest high
        return SO.eval_algorithm(close, low, high)



from speculator.features.OBV import OBV
import unittest

class OBVTest(unittest.TestCase):
    def test_eval_algorithm_case1(self):
        # Case 1: close > close_prev, return obv + volume
        curr = {'close': 5, 'volume': 1}
        prev = {'close': 4, 'obv': 2}
        self.assertEqual(OBV.eval_algorithm(curr, prev), 3)

    def test_eval_algorithm_case2(self):
        # Case 2: close == close_prev, return obv
        curr = {'close': 5, 'volume': 1}
        prev = {'close': 5, 'obv': 2}
        self.assertEqual(OBV.eval_algorithm(curr, prev), 2)

    def test_eval_algorithm_case3(self):
        # Case 3: close < close_prev, return obv - volume
        curr = {'close': 5, 'volume': 1}
        prev = {'close': 6, 'obv': 2}
        self.assertEqual(OBV.eval_algorithm(curr, prev), 1)



from pandas import crosstab
from sklearn.ensemble import RandomForestClassifier
from sklearn.metrics import accuracy_score

class RandomForest(RandomForestClassifier):
    """ Random Forest Classifier using a DataFrame from a Market instance

    Attributes:
        features: Pandas DataFrame, X axis of features
        targets: Pandas Series, Y axis of targets
    
    All other attributes and functions are defined
    within Scikit Learn's RandomForestClassifier class.
    """
    def __init__(self, features, targets, **kwargs):
        """ Inits a Random Forest Classifier with a market attribute

        Args:
            **kwargs: Scikit Learn's RandomForestClassifier kwargs
        """
        # Init model
        super().__init__(**kwargs)

        # Set axes
        self.features = features
        self.targets = targets

        # Train model
        self.fit(features.train, targets.train)

    def feature_importances(self):
        """ Return list of features and their importance in classification """
        feature_names = [feature for feature in self.features.train]
        return list(zip(feature_names, self.feature_importances_))

    def confusion_matrix(self, actual, preds):
        """ Confusion matrix of actual set to predicted set """
        return crosstab(actual, preds, rownames=['(A)'], colnames=['(P)'])

    def accuracy(self, x, y):
        """ Accuracy score of actual set to predicted set """
        return accuracy_score(y, self._predict_trends(x))

    def _predict_trends(self, *args, **kwargs):
        return super().predict(*args, **kwargs)

    def _predict_probas(self, *args, **kwargs):
        return super().predict_proba(*args, **kwargs)

    def _predict_logs(self, *args, **kwargs):
        return super().predict_log_proba(*args, **kwargs)




from setuptools import setup, find_packages
from codecs import open
from os import path

if path.exists('README.rst'):
    readme = open('README.rst').read()
elif path.exists('README.md'):
    readme = open('README.md').read()
else:
    readme = 'Failed to upload!'

"""
try:
    import pypandoc
    readme = pypandoc.convert(readme_path, 'rst')
except (IOError, ImportError):
    readme = open(readme_path).read()
"""

setup(
    name='speculator',
    version='1.0.2',
    description='Speculator predicts the price trend of cryptocurrencies like Bitcoin and Ethereum with machine learning models and technical analysis.',
    long_description=readme,
    url='https://github.com/amicks/Speculator',
    download_url='https://github.com/amicks/Speculator/archive/1.0.0.tar.gz',
    author='Allston Mickey',
    author_email='allston.mickey@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Flask',
        'Intended Audience :: Education',
        'Intended Audience :: Developers',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Topic :: Office/Business :: Financial :: Investment',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    keywords=[
        'finance',
        'price',
        'trend',
        'analysis',
        'machine learning',
        'artificial intelligence',
        'bitcoin',
        'ethereum',
        'cryptocurrency',
        'crypto',
        'api',
        'REST',
        'web'
    ],
    packages=find_packages(exclude=['docs', 'tests']),
    install_requires=[
        'delorean',
        'requests',
        'numpy',
        'scikit-learn',
        'tensorflow',
        'pandas',
        'flask',
        'flask-cache',
        'flask-restful',
        'webargs'
    ]
)



import numpy as np
import pandas as pd
from speculator import market
import tensorflow as tf
tf.logging.set_verbosity(tf.logging.ERROR)

class DeepNeuralNetwork(tf.estimator.DNNClassifier):
    def __init__(self, features, targets, **kwargs):
        feature_columns = [tf.feature_column.numeric_column(k) \
                           for k in features.train.columns]
        super().__init__(feature_columns=feature_columns,
                         hidden_units=[10, 20, 10],
                         n_classes=len(market.TARGET_CODES))

        # Set axes
        self.features = features
        self.targets = targets

        # Train model
        self.fit(features.train, y=targets.train)

    def fit(self, x, y=None):
        input_fn = self.get_input_fn(x,
                                     y=y,
                                     batch_size=128,
                                     queue_capacity=1000,
                                     shuffle=True,
                                     num_threads=1,
                                     num_epochs=None)
        self.train(input_fn=input_fn, steps=2000)

    def get_input_fn(self, x, y=None, shuffle=True, **kwargs):
        return tf.estimator.inputs.pandas_input_fn(x,
                                                   y=y,
                                                   shuffle=shuffle,
                                                   **kwargs)

    def accuracy(self, x, y):
        input_fn = self.get_input_fn(x, y)
        return self.evaluate(input_fn)['accuracy']

    def _predict_trends(self, x, y=None):
        input_fn = self.get_input_fn(x, y=y)
        # TF returns a dict generator, convert to list
        preds = list(self.predict(input_fn=input_fn))
        return [int(p['class_ids'][0]) for p in preds]
        
    def _predict_logs(self, x, y=None):
        input_fn = self.get_input_fn(x, y=y)
        # TF returns a dict generator, convert to list
        preds = list(self.predict(input_fn=input_fn))
        return [list(p['logits']) for p in preds]

    def _predict_probas(self, x, y=None):
        input_fn = self.get_input_fn(x, y=y)
        # TF returns a dict generator, convert to list
        preds = list(self.predict(input_fn=input_fn))
        return [list(p['probabilities']) for p in preds]
